import greenfoot.Actor;
import greenfoot.GreenfootImage;
import greenfoot.World;

import java.lang.Thread.State;
import java.util.List;

/**
 * class of Threadfoot
 * <p>
 * </p>
 * Counterpart of Greenfoot-class Actor. Working with Threadfoot you should use
 * always this class instead of class Actor.
 * <p>
 * </p>
 * Remark 1: If you override the method "addedToWorld" in a derived class you
 * have to call the ThreadActor-method "addedToWorld" within that method!
 * 
 * @author Dietrich Boles (Uni Oldenburg, Germany)
 * @version (2.0) 2010/12/08
 * 
 */
public abstract class ThreadActor extends Actor implements Runnable {

	private Thread thread;

	/**
	 * Creates a ThreadActor-Object with an internal Thread-Object
	 */
	public ThreadActor() {
		this.thread = new Thread(Threadfoot.getThreadGroup(), this);
	}

	/**
	 * This method does nothing. It should be overridden in subclasses to
	 * implement an thread's action.
	 */
	@Override
	public void run() {
	}

	/**
	 * Return the internal Thread-Object
	 * 
	 * @return the internal Thread-Object
	 */
	public Thread getThread() {
		return this.thread;
	}

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 * methods of Greenfoot-class Actor
	 */

	/**
	 * meaningless in Threadfoot
	 * 
	 * @see greenfoot.Actor#act()
	 */
	@Override
	public final void act() {
		// meaningless in Threadfoot
	}

	/**
	 * Return the x-coordinate of the object's current location. The value
	 * returned is the horizontal index of the object's cell in the world.
	 * 
	 * @return The x-coordinate of the object's current location.
	 * @throws IllegalStateException
	 *             If the actor has not been added into a world.
	 */
	@Override
	public int getX() throws IllegalStateException {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return super.getX();
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Return the y-coordinate of the object's current location. The value
	 * returned is the vertical index of the object's cell in the world.
	 * 
	 * @return The y-coordinate of the object's current location
	 * @throws IllegalStateException
	 *             If the actor has not been added into a world.
	 */
	@Override
	public int getY() {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return super.getY();
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Return the current rotation of the object. Rotation is expressed as a
	 * degree value, range (0..359). Zero degrees is to the east. The angle
	 * increases clockwise.
	 * 
	 * @see #setRotation(int)
	 * 
	 * @return The rotation in degrees.
	 */
	@Override
	public int getRotation() {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return super.getRotation();
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Set the rotation of the object. Rotation is expressed as a degree value,
	 * range (0..359). Zero degrees is to the east. The angle increases
	 * clockwise.
	 * 
	 * @param rotation
	 *            The rotation in degrees.
	 */
	@Override
	public void setRotation(int rotation) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				super.setRotation(rotation);
			}
			Threadfoot.repaintWorld();
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Assign a new location for this object. The location is specified as a
	 * cell index in the world.
	 * 
	 * If this method is overridden it is important to call this method with
	 * super.setLocation(x,y) from the overriding method.
	 * 
	 * @param x
	 *            Location index on the x-axis
	 * @param y
	 *            Location index on the y-axis
	 * @throws IllegalStateException
	 *             If the actor has not been added into a world.
	 */
	@Override
	public void setLocation(int x, int y) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				super.setLocation(x, y);
			}
			Threadfoot.repaintWorld();
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Return the world that this object lives in.
	 * 
	 * @return The world.
	 */
	@Override
	public World getWorld() {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return super.getWorld();
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Set the image for this object to the specified image.
	 * 
	 * @see #setImage(String)
	 * @param image
	 *            The image.
	 */
	@Override
	public void setImage(GreenfootImage image) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				super.setImage(image);
			}
			Threadfoot.repaintWorld();
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Returns the image used to represent this Actor. This image can be
	 * modified to change the object's appearance.
	 * 
	 * @return The object's image.
	 */
	@Override
	public GreenfootImage getImage() {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return super.getImage();
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Set an image for this object from an image file. The file may be in jpeg,
	 * gif or png format. The file should be located in the project directory.
	 * 
	 * @param filename
	 *            The name of the image file.
	 * @throws IllegalArgumentException
	 *             If the image can not be loaded.
	 */
	@Override
	public void setImage(String filename) throws IllegalArgumentException {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				super.setImage(new ThreadfootImage(filename));
			}
			Threadfoot.repaintWorld();
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Check whether this object intersects with another given object.
	 * 
	 * NOTE: When rotated, it only uses the axis alligned bounds of the rotated
	 * image.
	 * 
	 * @return True if the object's intersect, false otherwise.
	 */
	@Override
	protected boolean intersects(Actor other) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return super.intersects(other);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Return the neighbours to this object within a given distance. This method
	 * considers only logical location, ignoring extent of the image. Thus, it
	 * is most useful in scenarios where objects are contained in a single cell.
	 * <p>
	 * 
	 * All cells that can be reached in the number of steps given in 'distance'
	 * from this object are considered. Steps may be only in the four main
	 * directions, or may include diagonal steps, depending on the 'diagonal'
	 * parameter. Thus, a distance/diagonal specification of (1,false) will
	 * inspect four cells, (1,true) will inspect eight cells.
	 * <p>
	 * 
	 * @param distance
	 *            Distance (in cells) in which to look for other objects.
	 * @param diagonal
	 *            If true, include diagonal steps.
	 * @param cls
	 *            Class of objects to look for (passing 'null' will find all
	 *            objects).
	 * @return A list of all neighbours found.
	 */
	@SuppressWarnings("rawtypes")
	@Override
	protected List getNeighbours(int distance, boolean diagonal, Class cls) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return super.getNeighbours(distance, diagonal, cls);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Return all objects that intersect the given location (relative to this
	 * object's location). <br>
	 * 
	 * @return List of objects at the given offset. The list will include this
	 *         object, if the offset is zero.
	 * @param dx
	 *            X-coordinate relative to this objects location.
	 * @param dy
	 *            y-coordinate relative to this objects location.
	 * @param cls
	 *            Class of objects to look for (passing 'null' will find all
	 *            objects).
	 */
	@SuppressWarnings("rawtypes")
	@Override
	protected List getObjectsAtOffset(int dx, int dy, Class cls) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return super.getObjectsAtOffset(dx, dy, cls);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Return one object that is located at the specified cell (relative to this
	 * objects location). Objects found can be restricted to a specific class
	 * (and its subclasses) by supplying the 'cls' parameter. If more than one
	 * object of the specified class resides at that location, one of them will
	 * be chosen and returned.
	 * 
	 * @param dx
	 *            X-coordinate relative to this objects location.
	 * @param dy
	 *            y-coordinate relative to this objects location.
	 * @param cls
	 *            Class of objects to look for (passing 'null' will find all
	 *            objects).
	 * @return An object at the given location, or null if none found.
	 */
	@SuppressWarnings("rawtypes")
	@Override
	protected Actor getOneObjectAtOffset(int dx, int dy, Class cls) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return super.getOneObjectAtOffset(dx, dy, cls);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Return all objects within range 'radius' around this object. An object is
	 * within range if the distance between its centre and this object's centre
	 * is less than or equal to 'radius'.
	 * 
	 * @param radius
	 *            Radius of the circle (in cells)
	 * @param cls
	 *            Class of objects to look for (passing 'null' will find all
	 *            objects).
	 */
	@SuppressWarnings("rawtypes")
	@Override
	protected List getObjectsInRange(int radius, Class cls) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return super.getObjectsInRange(radius, cls);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Return all the objects that intersect this object. This takes the
	 * graphical extent of objects into consideration. <br>
	 * 
	 * NOTE: Does not take rotation into consideration.
	 * 
	 * @param cls
	 *            Class of objects to look for (passing 'null' will find all
	 *            objects).
	 */
	@SuppressWarnings("rawtypes")
	@Override
	protected List getIntersectingObjects(Class cls) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return super.getIntersectingObjects(cls);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Return an object that intersects this object. This takes the graphical
	 * extent of objects into consideration. <br>
	 * 
	 * NOTE: Does not take rotation into consideration.
	 * 
	 * @param cls
	 *            Class of objects to look for (passing 'null' will find all
	 *            objects).
	 */
	@SuppressWarnings("rawtypes")
	@Override
	protected Actor getOneIntersectingObject(Class cls) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return super.getOneIntersectingObject(cls);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 * internal methods
	 */

	/**
	 * This method is called by the Greenfoot system when the object has been
	 * inserted into the world. This method can be overridden to implement
	 * custom behaviour when the actor is inserted into the world.
	 * <p>
	 * </p>
	 * If you override this method in a derived class you have to call this
	 * method within that method!
	 * 
	 * @param world
	 *            The world the object was added to.
	 */
	@Override
	protected void addedToWorld(World world) {
		// synchronized (Threadfoot.getSyncObject()) {
		if (Threadfoot.isRunning()) {
			this.start();
		}
		// }
	}

	void start() {
		if (this.thread.getState() == State.NEW) {
			this.thread.start();
		}
	}

	@SuppressWarnings("deprecation")
	void stop() {
		this.thread.stop();
	}

}
