import greenfoot.Greenfoot;
import greenfoot.MouseInfo;
import greenfoot.World;

import java.util.List;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

/**
 * class of Threadfoot
 * <p>
 * </p>
 * Counterpart of Greenfoot-class Greenfoot. Working with Threadfoot you should
 * use always this class instead of class Greenfoot.
 * 
 * @author Dietrich Boles (Uni Oldenburg, Germany)
 * @version (2.0) 2010/12/08
 * 
 */
public class Threadfoot {

	final private static String GROUP_NAME = "Threadfoot";
	final private static ThreadGroup threadGroup = new ThreadGroup(
			Threadfoot.GROUP_NAME);
	final private static Object syncObject = new Object();
	private static Lock lock = new ReentrantLock(true);

	private static boolean running = false;
	private static boolean frozen = false;
	private static boolean resetted = false;
	private static World world = null;

	/**
	 * freezes the repainting of the world until unfreeze has been called
	 */
	public static void freeze() {
		Threadfoot.frozen = true;
	}

	/**
	 * counterpart of method freeze
	 */
	public static void unfreeze() {
		Threadfoot.frozen = false;
		Threadfoot.repaintWorld(); // repaint the world
	}

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 * methods of Greenfoot-class Greenfoot
	 */

	/**
	 * Get the most recently pressed key, since the last time this method was
	 * called. If no key was pressed since this method was last called, it will
	 * return null. If more than one key was pressed, this returns only the most
	 * recently pressed key.
	 * 
	 * @return The name of the most recently pressed key
	 */
	public static String getKey() {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return Greenfoot.getKey();
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Check whether a given key is currently pressed down.
	 * 
	 * @param keyName
	 *            The name of the key to check
	 * @return True if the key is down
	 */
	public static boolean isKeyDown(String keyName) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return Greenfoot.isKeyDown(keyName);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Delay the current execution by a number of time steps. The size of one
	 * time step is defined by the Greenfoot environment (the speed slider).
	 * 
	 * @see #setSpeed(int)
	 */
	public static void delay(int time) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				Greenfoot.delay(time);
			}
			Threadfoot.repaintWorld();
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Set the speed of the execution.
	 * 
	 * @param speed
	 *            The new speed. the value must be in the range (1..100)
	 */
	public static void setSpeed(int speed) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				Greenfoot.setSpeed(speed);
			}
			Threadfoot.repaintWorld();
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Pause the execution.
	 */
	public static void stop() {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				Greenfoot.stop();
			}
			Threadfoot.repaintWorld();
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Run (or resume) the execution.
	 */
	public static void start() {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				Greenfoot.start();
			}
			Threadfoot.repaintWorld();
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Return a random number between 0 (inclusive) and limit (exclusive).
	 */
	public static int getRandomNumber(int limit) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return Greenfoot.getRandomNumber(limit);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Play sound from a file. The following formats are supported: AIFF, AU and
	 * WAV.
	 * <p>
	 * 
	 * The file name may be an absolute path, a base name for a file located in
	 * the project directory or in the sounds directory of the project
	 * directory.
	 * 
	 * @param soundFile
	 *            Typically the name of a file in the sounds directory of the
	 *            scenario.
	 * @throws IllegalArgumentException
	 *             If the sound can not be loaded.
	 */
	public static void playSound(final String soundFile) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				Greenfoot.playSound(soundFile);
			}
			Threadfoot.repaintWorld();
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * True if the mouse has been pressed (changed from a non-pressed state to
	 * being pressed) on the given object. If the parameter is an Actor the
	 * method will only return true if the mouse has been pressed on the given
	 * actor. If there are several actors at the same place, only the top most
	 * actor will receive the press. If the parameter is a World then true will
	 * be returned if the mouse was pressed on the world background. If the
	 * parameter is null, then true will be returned for any mouse press,
	 * independent of the target pressed on.
	 * 
	 * @param obj
	 *            Typically one of Actor, World or null
	 * @return True if the mouse has been pressed as explained above
	 */
	public static boolean mousePressed(Object obj) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return Greenfoot.mousePressed(obj);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * True if the mouse has been clicked (pressed and released) on the given
	 * object. If the parameter is an Actor the method will only return true if
	 * the mouse has been clicked on the given actor. If there are several
	 * actors at the same place, only the top most actor will receive the click.
	 * If the parameter is a World then true will be returned if the mouse was
	 * clicked on the world background. If the parameter is null, then true will
	 * be returned for any click, independent of the target clicked on.
	 * 
	 * @param obj
	 *            Typically one of Actor, World or null
	 * @return True if the mouse has been clicked as explained above
	 */
	public static boolean mouseClicked(Object obj) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return Greenfoot.mouseClicked(obj);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * True if the mouse has been dragged on the given object. The mouse is
	 * considered to be dragged on an object if the drag started on that object
	 * - even if the mouse has since been moved outside of that object.
	 * <p>
	 * If the parameter is an Actor the method will only return true if the drag
	 * started on the given actor. If there are several actors at the same
	 * place, only the top most actor will receive the drag. If the parameter is
	 * a World then true will be returned if the drag action was started on the
	 * world background. If the parameter is null, then true will be returned
	 * for any drag action, independent of the target clicked on.
	 * 
	 * @param obj
	 *            Typically one of Actor, World or null
	 * @return True if the mouse has been dragged as explained above
	 */
	public static boolean mouseDragged(Object obj) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return Greenfoot.mouseDragged(obj);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * True if a mouse drag has ended. This happens when the mouse has been
	 * dragged and the mouse button released.
	 * <p>
	 * If the parameter is an Actor the method will only return true if the drag
	 * started on the given actor. If there are several actors at the same
	 * place, only the top most actor will receive the drag. If the parameter is
	 * a World then true will be returned if the drag action was started on the
	 * world background. If the parameter is null, then true will be returned
	 * for any drag action, independent of the target clicked on.
	 * 
	 * @param obj
	 *            Typically one of Actor, World or null
	 * @return True if the mouse has been dragged as explained above
	 */
	public static boolean mouseDragEnded(Object obj) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return Greenfoot.mouseDragEnded(obj);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * True if the mouse has been moved on the given object. The mouse is
	 * considered to be moved on an object if the mouse pointer is above that
	 * object.
	 * <p>
	 * If the parameter is an Actor the method will only return true if the move
	 * is on the given actor. If there are several actors at the same place,
	 * only the top most actor will receive the move. If the parameter is a
	 * World then true will be returned if the move was on the world background.
	 * If the parameter is null, then true will be returned for any move,
	 * independent of the target under the move location.
	 * 
	 * @param obj
	 *            Typically one of Actor, World or null
	 * @return True if the mouse has been moved as explained above
	 */
	public static boolean mouseMoved(Object obj) {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return Greenfoot.mouseMoved(obj);
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/**
	 * Return a mouse info object with information about the state of the mouse.
	 * 
	 * @return The info about the current state of the mouse. Or null if nothing
	 *         has happened with the mouse since it was last checked.
	 */
	public static MouseInfo getMouseInfo() {
		Threadfoot.lock();
		try {
			synchronized (Threadfoot.getSyncObject()) {
				return Greenfoot.getMouseInfo();
			}
		} finally {
			Threadfoot.unlock();
		}
	}

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 * internal methods
	 */


	@SuppressWarnings("rawtypes")
	static void startThreads(World world) {
		synchronized (Threadfoot.getSyncObject()) {
			if (!Threadfoot.running) {
				if (!Threadfoot.resetted) {
					SwingUtilities.invokeLater(new Runnable() {
						public void run() {
							JOptionPane
									.showMessageDialog(
											null,
											"You cannot start the scenario twice in a row.\nPress the \"Reset\"-button first!",
											"Information",
											JOptionPane.INFORMATION_MESSAGE);
							Greenfoot.stop();
						}
					});
				} else {
					Threadfoot.lock = new ReentrantLock(true);
					Threadfoot.world = world;
					Threadfoot.running = true;
					Threadfoot.frozen = false;
					Threadfoot.resetted = false;
					List threads = world.getObjects(ThreadActor.class);
					for (Object thread : threads) {
						((ThreadActor) thread).start();
					}
				}
			}
		}
	}

	@SuppressWarnings("rawtypes")
	static void stopThreads() {
		synchronized (Threadfoot.getSyncObject()) {
			if (Threadfoot.running) {
				List threads = Threadfoot.world.getObjects(ThreadActor.class);
				for (Object thread : threads) {
					((ThreadActor) thread).stop();
				}
				Threadfoot.running = false;
				Threadfoot.frozen = false;
			}
		}
	}

	static void setResetted() {
		Threadfoot.resetted = true;
	}

	static Object getSyncObject() {
		return Threadfoot.syncObject;
	}

	static ThreadGroup getThreadGroup() {
		return Threadfoot.threadGroup;
	}

	static void lock() {
		if (Threadfoot.isMemberOf(Thread.currentThread(),
				Threadfoot.threadGroup)) {
			Threadfoot.lock.lock();
		}
	}

	static void unlock() {
		if (Threadfoot.isMemberOf(Thread.currentThread(),
				Threadfoot.threadGroup)) {
			Threadfoot.lock.unlock();
			Thread.yield();
		}
	}

	static boolean isRunning() {
		return Threadfoot.running;
	}

	static void repaintWorld() {
		if (Threadfoot.running && !Threadfoot.frozen) {
			Greenfoot.delay(1);
		}
	}

	static boolean isMemberOf(Thread thread, ThreadGroup group) {
		return thread.getThreadGroup().getName().equals(Threadfoot.GROUP_NAME);
	}
}
